"""
This module is designed for desmostraing binary score plots
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import precision_recall_curve, roc_curve, auc


class BinaryScorePlot(object):
    """
    Class to construct plots to analyze performance of binary classifiers.
    Mainly acts as a wrapper for existing metrics and plotting functions.

    Input:
        - scores: 1D numpy array of model scores to plot (defaults to empty
          array)
        - labels: 1D numpy array of model labels to plot (defaults to empty
          array)
    """

    def __init__(self, scores=np.array([]), labels=np.array([]), threshold=.5):
        self._scores = scores
        self._labels = labels
        self._threshold = threshold

        pred_label = (scores > threshold) + 0
        cal_1 = pred_label + labels
        cal_2 = pred_label - labels
        df_curr = pd.DataFrame({'scores': scores, 'actual_label': labels,
                                'pred_label': pred_label, 'group': ''})
        df_curr.at[cal_1 == 2, 'group'] = 'TP'
        df_curr.at[cal_1 == 0, 'group'] = 'TN'
        df_curr.at[cal_2 == 1, 'group'] = 'FP'
        df_curr.at[cal_2 == -1, 'group'] = 'FN'
        df_curr['position'] = df_curr['actual_label'] + \
            np.random.uniform(low=-0.3, high=0.3, size=len(df_curr))
        self._df = df_curr
        sns.set_style("darkgrid")

    def get_scores(self):
        """
        Get model scores to plot
        Output:
            - 1D numpy array of model scores
        """

        return self._scores

    def get_labels(self):
        """
        Get model labels to plot
        Output:
            - 1D numpy array of model labels
        """

        return self._labels

    def get_df(self):
        """
        Get data for scatterplot
        Output:
            - pandas dataframe
        """

        return self._df

    def get_threshold(self):
        """
        Output:
            - threshold
        """

        return self._threshold

    def set_scores(self, scores):
        """
        Set model scores to plot
        Input:
            - 1D numpy array of model scores
        """

        self._scores = scores

    def set_labels(self, labels):
        """
        Set model labels to plot
        Input:
            - 1D numpy array of model labels
        """

        self._labels = labels

    def plot_hist(self, bins=30):
        """
        Plot two histograms: one of
        the actual binary labels
        and one of the model scores
        Input:
            - number of histogram bins to use
        """

        labels = self.get_labels()
        scores = self.get_scores()

        plt.clf()
        plt.subplot(221)
        sns.distplot(labels, kde=False)
        plt.xlabel('Actual Label', fontsize=12)
        plt.xticks([0, 1])
        plt.subplot(222)
        sns.distplot(scores, bins=bins, kde=False)
        plt.xlabel('Scores', fontsize=12)
        plt.xticks([0, 0.25, 0.5, 0.75, 1])
        plt.subplots_adjust(hspace=0.5)
        plt.subplot(223)
        sns.distplot(scores[labels == 0], bins=bins, kde=False)
        plt.xlabel('Actual Label = 0', fontsize=12)
        plt.subplot(224)
        sns.distplot(scores[labels == 1], bins=bins, kde=False)
        plt.xlabel('Actual Label = 1', fontsize=12)
        plt.suptitle('Histograms of Model Scores by Actual Label', fontsize=16)
        plt.show()

    def plot_jitter(self):
        """
        Make jitter plot
        """

        df_curr = self.get_df()
        threshold = self.get_threshold()

        plt.clf()
        sns.scatterplot(x='position', y='scores', hue='group', s=10, alpha=0.8,
                        data=df_curr)
        plt.hlines(y=threshold, xmin=-0.3, xmax=1.8, color='red')
        plt.xticks([0, 1])
        plt.xlabel('Actual label', fontsize=14)
        plt.ylabel('Scores', fontsize=14)
        title = 'Scatterplot of Model Scores with Threshold = '
        title += str(threshold)
        plt.suptitle(title, fontsize=16)
        plt.show()

    def plot_pr_by_threshold(self):
        """
        Plots model precision and recall
        by threshold, using matplotlib and seaborn
        allowing a user to visualize model performance
        at various thresholds
        """

        labels = self.get_labels()
        scores = self.get_scores()

        precision, recall, thresholds = precision_recall_curve(labels, scores)
        thresholds = np.append(thresholds, 1)

        plt.clf()
        plt.plot(thresholds, precision, color=sns.color_palette()[0])
        plt.plot(thresholds, recall, color=sns.color_palette()[1])

        leg = plt.legend(('Precision', 'Recall'), frameon=True)
        leg.get_frame().set_edgecolor('k')

        plt.xlabel('Threshold', fontsize=14)
        plt.ylabel('Percent', fontsize=14)
        plt.suptitle('Precision and Recall by Model Threshold', fontsize=16)

        plt.show()

    def plot_roc(self):
        """
        Plots the true positive rate vs. the false positive rate
        """

        labels = self.get_labels()
        scores = self.get_scores()
        fpr, tpr, thresholds = roc_curve(labels, scores)
        roc_auc = auc(fpr, tpr)

        plt.clf()
        plt.plot(fpr, tpr, color='darkgreen', lw=2,
                 label='AUC = %0.3f' % (roc_auc))
        plt.plot([0, 1], [0, 1], color='red', lw=2, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate', fontsize=14)
        plt.ylabel('True Positive Rate', fontsize=14)
        plt.suptitle('Receiver Operating Characteristic', fontsize=16)
        plt.legend(loc="lower right")

        plt.show()

    def optimal_threshold(self, by_mode='roc'):
        """
        Calculate the optimal threshold by auc of either ROC or PR curve
        Input:
            - 1D numpy array of model scores
            - 1D numpy array of actual labels
            - Curve to use
        """

        labels = self.get_labels()
        scores = self.get_scores()

        if by_mode == 'roc':
            fpr, tpr, thresholds = roc_curve(labels, scores)
            tnr = 1 - fpr
            t_f = tpr - tnr
            optimal_threshold = round(thresholds[abs(t_f).argsort()[0]], 2)
        else:
            precision, recall, thresholds = precision_recall_curve(labels,
                                                                   scores)
            thresholds = np.append(thresholds, 1)
            t_f = precision - recall
            optimal_threshold = round(thresholds[abs(t_f).argsort()[0]], 2)
        return optimal_threshold



"""
A basic dashboard to present linear regression model assumptions diagnostics
To run: python regression_diagnostics.py
"""

import pandas as pd
import numpy as np
import plotly.graph_objs as go
import statsmodels.api as sm
from statsmodels.graphics.gofplots import ProbPlot

import dash
from dash.dependencies import Input, Output
import dash_core_components as dcc
import dash_html_components as html

import regression_analysis_plot as plt

EXTERNAL_STYLESHEETS = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
APP = dash.Dash(__name__, external_stylesheets=EXTERNAL_STYLESHEETS)
LOWESS = sm.nonparametric.lowess

# Read csv dataset
OPTION = 'Please enter a CSV data file url or local path:\n'
OPTION += 'Url example: www.someplace.com/mydata.csv\n'
OPTION += 'Local path example: ./mydata.csv\n'
ADDRESS = input(OPTION)
DF = pd.read_csv(ADDRESS, sep=',')
DF = DF.dropna()  # remove missing data
DF = DF.select_dtypes(exclude=['object'])  # keep only numeric columns
COLS = DF.columns
MYCLASS = plt.RegressionAnalysisPlot(DF)

# Layout and Plots
APP.layout = html.Div([
    html.H1('Linear Regression Model Diagnostics',
            style={'textAlign': 'center'}),
    html.Div('Data Source: %s\n' % ADDRESS, style={'font-weight': 'bold'}),
    html.Div([
        html.Div([
            dcc.Dropdown(
                id='predictors',
                options=[{'label': i, 'value': i} for i in COLS],
                placeholder='Select predictor variable(s)',
                multi=True
            )
        ], style={'width': '49%', 'display': 'inline-block'}),
        html.Div([
            dcc.Dropdown(
                id='response',
                options=[{'label': i, 'value': i} for i in COLS],
                placeholder='Select a response variable'
            )
        ], style={'width': '49%', 'display': 'inline-block'})
    ], style={
        'borderBottom': 'thin lightgrey solid',
        'backgroundColor': 'rgb(250, 250, 250)',
        'padding': '10px 5px'}),

    html.Div([
        dcc.Graph(id='residual-plot')
    ], style={'width': '49%', 'display': 'inline-block'}),

    html.Div([
        dcc.Graph(id='qq-plot')
    ], style={'width': '49%', 'display': 'inline-block'}),

    html.Div([
        dcc.Graph(id='scale-location-plot')
    ], style={'width': '49%', 'display': 'inline-block'}),

    html.Div([
        dcc.Graph(id='residual-leverage-plot')
    ], style={'width': '49%', 'display': 'inline-block'}),

    html.Div('Residuals versus Predicted Plot:',
             style={'font-weight': 'bold'}),
    html.Div('Checking the assumption of linearity and homoscedasticity. \
             If the residuals appear to be very large (big positive value \
             or big negative value), the model does not meet the linear \
             model assumption. To assess the assumption of linearity we \
             want to ensure that the residuals are not too far away from 0.'),
    html.Div('Normal Q-Q Plot:', style={'font-weight': 'bold'}),
    html.Div('Checking the normality assumption by comparing the residuals \
             to "ideal" normal observations. If observations lie well \
             along the 45-degree line in the QQ-plot, we may assume that \
             normality holds here.'),
    html.Div('Scale-Location Plot:', style={'font-weight': 'bold'}),
    html.Div('Checking the assumption of homoscedasticity(equal variance). \
             If residuals are spread equally along the ranges of \
             predictors, the assumption of equal variance (homoscedasticity) \
             is satisfied. It’s good if you see a horizontal line with \
             equally (randomly) spread points.'),
    html.Div('Residuals vs Leverage Plot:', style={'font-weight': 'bold'}),
    html.Div('Checking if there are influential cases. Not all outliers are \
             influential in linear regression analysis, we watch out for \
             outlying values at the upper right corner or at the lower \
             right corner, since the regression results will be altered \
             if we exclude those cases.')
])


@APP.callback(
    Output('residual-plot', 'figure'),
    [Input('predictors', 'value'), Input('response', 'value')])
def update_resid_plot(predictors, response):
    """
    Residual Plot
    """
    MYCLASS.reg(var_x=predictors, var_y=response)
    model = MYCLASS.get_model()
    fitted = model.fittedvalues
    resid = model.resid
    smooth = LOWESS(resid, fitted)
    marker_size = 8
    opacity = 0.5
    traces = [
        go.Scatter(
            x=fitted,
            y=resid,
            mode='markers',
            name='data',
            marker=dict(
                size=marker_size,
                opacity=opacity
            )
        ),
        go.Scatter(
            x=smooth[:, 0],
            y=smooth[:, 1],
            mode='lines',
            name='smoother',
            line=dict(
                width=1,
                color='red'
            )
        )
    ]

    return {
        'data': traces,
        'layout': dict(
            title='Residuals vs Fitted',
            xaxis={'title': 'Fitted Values'},
            yaxis={'title': 'Residuals'},
            plot_bgcolor='#e6e6e6',
            showlegend=False,
            hovermode='closest'
        )
    }


@APP.callback(
    Output('qq-plot', 'figure'),
    [Input('predictors', 'value'), Input('response', 'value')])
def update_qq_plot(predictors, response):
    """
    Normal QQ Plot
    """
    MYCLASS.reg(predictors, response)
    model = MYCLASS.get_model()
    resid_norm = model.get_influence().resid_studentized_internal
    qq_res = ProbPlot(resid_norm)
    theo = qq_res.theoretical_quantiles
    sample = qq_res.sample_quantiles
    marker_size = 8
    opacity = 0.5
    traces = [
        go.Scatter(
            x=theo,
            y=sample,
            mode='markers',
            name='data',
            marker=dict(
                size=marker_size,
                opacity=opacity
            )
        ),
        go.Scatter(
            x=theo,
            y=theo,
            type='scatter',
            mode='lines',
            name='line',
            line=dict(
                width=1,
                color='red'
            )
        )
    ]

    return {
        'data': traces,
        'layout': dict(
            title='Normal Q-Q',
            xaxis={'title': 'Theoretical Quantiles'},
            yaxis={'title': 'Standardized Residuals'},
            plot_bgcolor='#e6e6e6',
            showlegend=False,
            hovermode='closest'
        )
    }


@APP.callback(
    Output('scale-location-plot', 'figure'),
    [Input('predictors', 'value'), Input('response', 'value')])
def update_scale_loc_plot(predictors, response):
    """
    Scale-Location Plot
    """
    MYCLASS.reg(predictors, response)
    model = MYCLASS.get_model()
    fitted = model.fittedvalues
    resid_norm = model.get_influence().resid_studentized_internal
    resid_norm_abs_sqrt = np.sqrt(np.abs(resid_norm))
    smooth = LOWESS(resid_norm_abs_sqrt, fitted)
    marker_size = 8
    opacity = 0.5
    traces = [
        go.Scatter(
            x=fitted,
            y=resid_norm_abs_sqrt,
            mode='markers',
            name='data',
            marker=dict(
                size=marker_size,
                opacity=opacity
            )
        ),
        go.Scatter(
            x=smooth[:, 0],
            y=smooth[:, 1],
            mode='lines',
            name='smoother',
            line=dict(
                width=1,
                color='red'
            )
        )
    ]

    return {
        'data': traces,
        'layout': dict(
            title='Scale Location',
            xaxis={'title': 'Fitted values'},
            yaxis={'title': 'Absolute Squared Normalized Residuals'},
            plot_bgcolor='#e6e6e6',
            showlegend=False,
            hovermode='closest'
        )
    }


@APP.callback(
    Output('residual-leverage-plot', 'figure'),
    [Input('predictors', 'value'), Input('response', 'value')])
def update_resid_lever_plot(predictors, response):
    """
    Residual vs Leverage Plot
    """
    MYCLASS.reg(predictors, response)
    model = MYCLASS.get_model()
    model_leverage = model.get_influence().hat_matrix_diag
    resid_norm = model.get_influence().resid_studentized_internal
    smooth = LOWESS(resid_norm, model_leverage)
    marker_size = 8
    opacity = 0.5
    traces = [
        go.Scatter(
            x=model_leverage,
            y=resid_norm,
            mode='markers',
            name='data',
            marker=dict(
                size=marker_size,
                opacity=opacity
            )
        ),
        go.Scatter(
            x=smooth[:, 0],
            y=smooth[:, 1],
            mode='lines',
            name='smoother',
            line=dict(
                width=1,
                color='red'
            )
        )
    ]

    return {
        'data': traces,
        'layout': dict(
            title='Residuals vs Leverage',
            xaxis={'title': 'Leverage'},
            yaxis={'title': 'Standardized Residuals'},
            plot_bgcolor='#e6e6e6',
            showlegend=False,
            hovermode='closest'
        )
    }


if __name__ == '__main__':
    APP.run_server(debug=False)



"""
Unit tests for analysis_plot.py
"""

import unittest
from unittest.mock import patch
import seaborn as sns
import regression_analysis_plot as plt


DF = sns.load_dataset('iris')
PREDS = [DF.columns[0]]
RESP = DF.columns[1]
MYPLOT = plt.RegressionAnalysisPlot(DF, predictors=PREDS, response=RESP)


class TestRegressionAnalysisPlot(unittest.TestCase):
    """
    32 tests for analysis_plot
    """

    def test_init_predictors(self):
        """
        Check constructor error if predictors not chosen from dataframe
        """
        try:
            test_preds = PREDS + ['somename']
            plt.RegressionAnalysisPlot(DF, predictors=test_preds)
        except Exception as err:
            msg = 'Input predictor variable(s) not existed '
            msg += 'in the given dataframe'
            self.assertEqual(err.args[0], msg)

    def test_init_response(self):
        """
        Check constructor error if response not chosen from dataframe
        """
        try:
            test_resp = 'somename'
            plt.RegressionAnalysisPlot(DF, response=test_resp)
        except Exception as err:
            msg = 'Input response variable not existed in the '
            msg += 'given dataframe'
            self.assertEqual(err.args[0], msg)

    def test_get_dataframe(self):
        """
        Test if dataframe matched
        """
        self.assertEqual(MYPLOT.get_dataframe().all().all(),
                         DF.all().all(), "Dataframe not matched")

    def test_get_predictors(self):
        """
        Test if predictors matched and in dataframe columns
        """
        self.assertEqual(MYPLOT.get_predictors(), PREDS,
                         "Predictors not matched")
        for pred_my in MYPLOT.get_predictors():
            self.assertIn(pred_my, DF.columns,
                          "Predictors not in dataframe columns")

    def test_get_response(self):
        """
        Test if response matched and in dataframe columns
        """
        self.assertIn(MYPLOT.get_response(), RESP,
                      "Predictors not matched")
        self.assertIn(MYPLOT.get_response(), DF.columns,
                      "Predictors not in dataframe columns")

    def test_set_predictors(self):
        """
        Test predictors matched
        """
        test_preds = [DF.columns[1]]
        MYPLOT.set_predictors(test_preds)
        self.assertEqual(MYPLOT.get_predictors(), test_preds,
                         "Predictors not matched")

    def test_set_response(self):
        """
        Test response matched
        """
        test_resp = DF.columns[2]
        MYPLOT.set_response(test_resp)
        self.assertEqual(MYPLOT.get_response(), test_resp,
                         "Response not matched")

    def test_set_predictors_err(self):
        """
        Check set error if predictors not chosen from dataframe
        """
        try:
            test_preds = PREDS + ['somename']
            MYPLOT.set_predictors(test_preds)
        except Exception as err:
            msg = 'Input predictor variable(s) not existed in the '
            msg += 'given dataframe'
            self.assertEqual(err.args[0], msg)

    def test_set_response_err(self):
        """
        Check set error if response not chosen from dataframe
        """
        try:
            test_resp = 'somename'
            MYPLOT.set_response(test_resp)
        except Exception as err:
            msg = 'Input response variable not existed in the given '
            msg += 'dataframe'
            self.assertEqual(err.args[0], msg)

    def test_matrix_plot_label(self):
        """
        Test matrix_plot with assigned label
        """
        MYPLOT.matrix_plot(label='species')
        self.assertEqual(1, 1)

    def test_matrix_plot_no_label(self):
        """
        Test matrix_plot without assigned label
        """
        MYPLOT.matrix_plot()
        self.assertEqual(1, 1)

    def test_matrix_plot_no_huelabel(self):
        """
        Test matrix_plot without showing huelabel
        """
        TESTPLOT = plt.RegressionAnalysisPlot(DF, predictors=PREDS)
        TESTPLOT.matrix_plot()
        self.assertEqual(1, 1)

    def test_corr_heatmap_figsize(self):
        """
        Test corr_heatmap with assigned figsize
        """
        MYPLOT.corr_heatmap(figsize=(10, 15))
        self.assertEqual(1, 1)

    def test_corr_heatmap_no_figsize(self):
        """
        Test corr_heatmap without assigned figsize
        """
        MYPLOT.corr_heatmap()
        self.assertEqual(1, 1)

    def test_reg_plot(self):
        """
        Test reg_plot
        """
        MYPLOT.reg_plot(DF.columns[0], DF.columns[1])
        self.assertEqual(1, 1)

    def test_box_plot(self):
        """
        Test box_plot
        """
        MYPLOT.box_plot(DF.columns[0], DF.columns[1])
        self.assertEqual(1, 1)

    def test_dist_plot(self):
        """
        Test dist_plot
        """
        MYPLOT.dist_plot(DF.columns[0], 'species')
        self.assertEqual(1, 1)

    def test_reg_print_report(self):
        """
        Test linear regression model, no input, print report summary
        """
        with patch('builtins.print'):
            MYPLOT.reg(report=True)
            # string is the first line of report with no space
            string = str(MYPLOT.get_model().summary())[:78].replace(' ',
                                                                    '')
            self.assertEqual(string, 'OLSRegressionResults')

    def test_reg_assign_arg(self):
        """
        Test linear regression model, with input
        """
        MYPLOT.reg(var_x=DF.columns[0], var_y=DF.columns[1])
        string = str(MYPLOT.get_model().summary())
        # assigned arguments should take priority in model fitting
        self.assertIn(str(DF.columns[0]), string)
        self.assertIn(str(DF.columns[1]), string)

    def test_reg_arg_err(self):
        """
        Check error for linear regression model, with input not being casted
        """
        try:
            MYPLOT.reg(var_x=DF.columns[0], var_y='species')
        except Exception as err:
            msg = 'Predictor/Response data type cannot be casted. '
            msg += 'Please select again'
            self.assertEqual(err.args[0], msg)

    def test_resid_plot(self):
        """
        Test resid_plot
        """
        MYPLOT.resid_plot()
        self.assertEqual(1, 1)

    def test_resid_plot_assign_arg(self):
        """
        Test resid_plot, with input
        """
        MYPLOT.resid_plot(var_x=DF.columns[0], var_y=DF.columns[1])
        self.assertEqual(1, 1)

    def test_resid_plot_arg_err(self):
        """
        Check error for resid_plot, with no assigned variables
        """
        try:
            test_plot = plt.RegressionAnalysisPlot(DF)
            test_plot.resid_plot()
        except Exception as err:
            self.assertEqual(err.args[0],
                             'No predictors or response assigned')

    def test_qq_plot(self):
        """
        Test qq_plot
        """
        MYPLOT.qq_plot()
        self.assertEqual(1, 1)

    def test_qq_plot_assign_arg(self):
        """
        Test qq_plot, with input
        """
        MYPLOT.qq_plot(var_x=DF.columns[0], var_y=DF.columns[1])
        self.assertEqual(1, 1)

    def test_qq_plot_arg_err(self):
        """
        Check error for qq_plot, with no assigned variables
        """
        try:
            test_plot = plt.RegressionAnalysisPlot(DF)
            test_plot.qq_plot()
        except Exception as err:
            self.assertEqual(err.args[0],
                             'No predictors or response assigned')

    def test_scale_loc_plot(self):
        """
        Test scale_loc_plot
        """
        MYPLOT.scale_loc_plot()
        self.assertEqual(1, 1)

    def test_scale_loc_plot_assign_arg(self):
        """
        Test scale_loc_plot, with input
        """
        MYPLOT.scale_loc_plot(var_x=DF.columns[0], var_y=DF.columns[1])
        self.assertEqual(1, 1)

    def test_scale_loc_plot_arg_err(self):
        """
        Check error for scale_loc_plot, with no assigned variables
        """
        try:
            test_plot = plt.RegressionAnalysisPlot(DF)
            test_plot.scale_loc_plot()
        except Exception as err:
            self.assertEqual(err.args[0],
                             'No predictors or response assigned')

    def test_resid_lever_plot(self):
        """
        Test resid_lever_plot
        """
        MYPLOT.resid_lever_plot()
        self.assertEqual(1, 1)

    def test_resid_lever_plot_assign_arg(self):
        """
        Test resid_lever_plot, with input
        """
        MYPLOT.resid_lever_plot(var_x=DF.columns[0], var_y=DF.columns[1])
        self.assertEqual(1, 1)

    def test_resid_lever_plot_arg_err(self):
        """
        Check error for resid_lever_plot, with no assigned variables
        """
        try:
            test_plot = plt.RegressionAnalysisPlot(DF)
            test_plot.resid_lever_plot()
        except Exception as err:
            self.assertEqual(err.args[0],
                             'No predictors or response assigned')


if __name__ == "__main__":
    SUITE = unittest.TestLoader().loadTestsFromTestCase(
        TestRegressionAnalysisPlot)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



"""
Unit tests for clustering.py
"""

import unittest
from sklearn.datasets.samples_generator import make_blobs
import clustering_evaluate


class TestKmeansEval(unittest.TestCase):
    """
    Testing the function to run k-means++ clustering algo
    """

    def test_kmeans(self):
        """
        Testing the function to run k-means++ clustering algo for
        various number of clusters on the given input_features
        """
        selected_n_clusters = 10
        x_train, _y_train = make_blobs(n_samples=10000,
                                       centers=selected_n_clusters,
                                       n_features=300,
                                       cluster_std=0.001,
                                       random_state=0)
        opt_clusters = clustering_evaluate.kmeans_viz(
            x_train, plot_figure=False)
        self.assertTrue(selected_n_clusters == opt_clusters)

    def test_kmeans_figure(self):
        """
            Testing the function to run k-means++ clustering algo for
            various number of clusters on the given input_features
        """
        selected_n_clusters = 10
        x_train, _y_train = make_blobs(n_samples=10000,
                                       centers=selected_n_clusters,
                                       n_features=300,
                                       cluster_std=0.001,
                                       random_state=0)
        plt_clust, opt_clusters = clustering_evaluate.kmeans_viz(
            x_train, plot_figure=True)
        self.assertTrue(selected_n_clusters == opt_clusters)


if __name__ == "__main__":
    SUITE = unittest.TestLoader().loadTestsFromTestCase(TestKmeansEval)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



"""
Unit tests for binary_score_plot.py
"""

import unittest
import numpy as np
from binary_score_plot import BinaryScorePlot


# Test scores and labels, size = 20
SCORES = np.array([0.54792232, 0.96933133, 0.99987806, 0.71340985, 0.71342704,
                   0.61219246, 0.72836524, 0.61848749, 0.99380689, 0.81295423,
                   0.88095189, 0.48289547, 0.54198686, 0.80857878, 0.9975807,
                   0.93301802, 0.4716791, 0.96753287, 0.6641115, 0.9989481])
LABELS = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])


class TestBinaryScorePlot(unittest.TestCase):
    """
    Testing the plots in the binary_score_plot module
    """

    def test_threshold_assignment(self):
        """
        Testing the threshold assignment in constructor
        """
        threshold = 0.55

        bsp = BinaryScorePlot(SCORES, LABELS, threshold)
        assigned_threshold = bsp.get_threshold()
        self.assertTrue(threshold == assigned_threshold)

    def test_set_scores(self):
        """
        Test scores assignment
        """
        bsp = BinaryScorePlot(SCORES, LABELS)
        test_scores = np.zeros(20)
        bsp.set_scores(test_scores)
        self.assertEqual(bsp.get_scores().all(), test_scores.all())

    def test_set_labels(self):
        """
        Test labels assignment
        """
        bsp = BinaryScorePlot(SCORES, LABELS)
        test_labels = np.zeros(20)
        bsp.set_labels(test_labels)
        self.assertEqual(bsp.get_labels().all(), test_labels.all())

    def test_plots(self):
        """
        Test four plots
        """
        bsp = BinaryScorePlot(SCORES, LABELS)
        bsp.plot_hist()
        bsp.plot_jitter()
        bsp.plot_pr_by_threshold()
        bsp.plot_roc()
        self.assertEqual(1, 1)

    def test_optimal_threshold_range_roc(self):
        """
        Testing the optimal threshold range by roc (default)
        """
        bsp = BinaryScorePlot(SCORES, LABELS)
        opt_thresh = bsp.optimal_threshold(by_mode='roc')
        self.assertTrue(opt_thresh >= 0)
        self.assertTrue(opt_thresh <= 1)

    def test_optimal_threshold_range(self):
        """
            Testing the optimal threshold range not by roc
            """
        bsp = BinaryScorePlot(SCORES, LABELS)
        opt_thresh = bsp.optimal_threshold(by_mode='other')
        self.assertTrue(opt_thresh >= 0)
        self.assertTrue(opt_thresh <= 1)


if __name__ == "__main__":
    SUITE = unittest.TestLoader().loadTestsFromTestCase(TestBinaryScorePlot)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



"""
__init__
"""

from betas.binary_score_plot import BinaryScorePlot
from betas.regression_analysis_plot import RegressionAnalysisPlot
from betas import pca_evaluate
from betas import clustering_evaluate



"""
A basic dashboard to explore binary score plots
To run: bokeh serve --show binary_score_diagnostics.py
"""

import numpy as np
import pandas as pd
from os.path import dirname, join
from sklearn.metrics import precision_recall_curve, roc_curve, auc
from bokeh.models import ColumnDataSource, Legend, Slider, Label, CustomJS
from bokeh.models.tickers import FixedTicker
from bokeh.models.widgets import Button, PreText, TextInput
from bokeh.plotting import figure, curdoc
from bokeh.layouts import gridplot, column


# Scatterplot
scores = np.concatenate((np.random.uniform(low=0.9, high=1, size=100),
                         np.random.uniform(low=0.8, high=0.9, size=50),
                         np.random.uniform(low=0.7, high=0.8, size=25),
                         np.random.uniform(low=0.5, high=0.7, size=5),
                         np.random.uniform(low=0, high=0.1, size=100),
                         np.random.uniform(low=0.1, high=0.2, size=50),
                         np.random.uniform(low=0.2, high=0.3, size=25),
                         np.random.uniform(low=0.3, high=0.5, size=5),
                         np.random.uniform(low=0.5, high=1, size=20),
                         np.random.uniform(low=0, high=0.5, size=20)))
labels = np.concatenate((np.ones(180), np.zeros(180),
                         np.zeros(20), np.ones(20)))


def classify(scores, labels, threshold=0.5):
    pred_label = (scores > threshold) + 0
    cal_1 = pred_label + labels
    cal_2 = pred_label - labels
    result = pd.DataFrame({'scores': scores, 'actual_label': labels,
                           'pred_label': pred_label, 'group': ''})
    result.at[cal_1 == 2, 'group'] = 'TP'
    result.at[cal_1 == 0, 'group'] = 'TN'
    result.at[cal_2 == 1, 'group'] = 'FP'
    result.at[cal_2 == -1, 'group'] = 'FN'
    np.random.seed(0)
    noise = np.random.uniform(low=-0.3, high=0.3, size=len(result))
    result['position'] = result['actual_label'] + noise
    return result


target = classify(scores, labels, 0.65)

# Scatterplot
hline = ColumnDataSource(data=dict(x=[-0.3, 1.3],
                         y=[0.7, 0.7]))
TP = ColumnDataSource(data=dict(x=target['position'][target.group == 'TP'],
                                y=target['scores'][target.group == 'TP']))
TN = ColumnDataSource(data=dict(x=target['position'][target.group == 'TN'],
                                y=target['scores'][target.group == 'TN']))
FP = ColumnDataSource(data=dict(x=target['position'][target.group == 'FP'],
                                y=target['scores'][target.group == 'FP']))
FN = ColumnDataSource(data=dict(x=target['position'][target.group == 'FN'],
                                y=target['scores'][target.group == 'FN']))

p_scatter = figure(plot_width=380, plot_height=260,
                   title='Scatterplot of Model Scores')
r0 = p_scatter.circle('x', 'y', source=TP, size=3, color='navy')
r1 = p_scatter.circle('x', 'y', source=TN, size=3, color='green')
r2 = p_scatter.circle('x', 'y', source=FP, size=3, color='orange')
r3 = p_scatter.circle('x', 'y', source=FN, size=3, color='tomato')
p_scatter.line('x', 'y', source=hline, line_width=1.5, color='red',
               line_dash='dotdash')
p_scatter.xaxis.axis_label = 'Actual Label'
p_scatter.yaxis.axis_label = 'Scores'
p_scatter.xaxis.ticker = FixedTicker(ticks=[0, 1])
p_scatter.yaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_scatter.title.text_font_size = "20px"
legend = Legend(items=[('TP', [r0]), ('TN', [r1]), ('FP', [r2]), ('FN', [r3])],
                location="center")
p_scatter.add_layout(legend, 'right')


# ROC curve
fpr, tpr, thresholds = roc_curve(target.actual_label, target.scores)
roc_auc = auc(fpr, tpr)
threshold = 0.7
diff = thresholds - threshold
x_coord = fpr[abs(diff).argsort()[0]]
y_coord = tpr[abs(diff).argsort()[0]]
roc_vline = ColumnDataSource(data=dict(x=[x_coord, x_coord], y=[0, y_coord]))
roc_hline = ColumnDataSource(data=dict(x=[x_coord, 1], y=[y_coord, y_coord]))
roc = ColumnDataSource(data=dict(x=fpr, y=tpr))

p_roc = figure(plot_width=380, plot_height=260,
               title='ROC Curve of Model Scores')
p_roc.line('x', 'y', source=roc, line_width=2)
p_roc.line('x', 'y', source=roc_vline, line_width=1.5, color='red',
           line_dash='dotdash')
p_roc.line('x', 'y', source=roc_hline, line_width=1.5, color='red',
           line_dash='dotdash')
label_tpr = Label(x=0.4, y=0.3, text='TPR: ' + str(
                  round(y_coord, 3)), text_font_size='10pt')
label_tnr = Label(x=0.4, y=0.2, text='TNR: ' + str(
                  round(1 - x_coord, 3)), text_font_size='10pt')
p_roc.xaxis.axis_label = 'False Positive Rate'
p_roc.yaxis.axis_label = 'True Positive Rate'
p_roc.xaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_roc.yaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_roc.title.text_font_size = "20px"
p_roc.add_layout(label_tpr)
p_roc.add_layout(label_tnr)


# Precision and recall curve
precision, recall, thresholds_pr = precision_recall_curve(labels, scores)
thresholds_pr = np.append(thresholds_pr, 1)
diff_pr = thresholds_pr - 0.7
x_coord_pr = recall[abs(diff_pr).argsort()[0]]
y_coord_pr = precision[abs(diff_pr).argsort()[0]]
pr_vline = ColumnDataSource(data=dict(x=[x_coord_pr, x_coord_pr],
                                      y=[0, y_coord_pr]))
pr_hline = ColumnDataSource(data=dict(x=[0, x_coord_pr],
                                      y=[y_coord_pr, y_coord_pr]))
pr = ColumnDataSource(data=dict(x=recall, y=precision))

p_pr = figure(plot_width=380, plot_height=260,
              title='Precision and Recall Curve')
p_pr.line('x', 'y', source=pr, line_width=2)
p_pr.line('x', 'y', source=pr_vline, line_width=1.5, color='red',
          line_dash='dotdash')
p_pr.line('x', 'y', source=pr_hline, line_width=1.5, color='red',
          line_dash='dotdash')
label_pre = Label(x=0.4, y=0.3, text='Precision: ' + str(
                  round(y_coord_pr, 3)), text_font_size='10pt')
label_rec = Label(x=0.4, y=0.2, text='Recall: ' + str(
                  round(x_coord_pr, 3)), text_font_size='10pt')
p_pr.xaxis.axis_label = 'Recall'
p_pr.yaxis.axis_label = 'Precision'
p_pr.xaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_pr.yaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_pr.title.text_font_size = "20px"
p_pr.add_layout(label_pre)
p_pr.add_layout(label_rec)


# Bar plot
bar_data = ColumnDataSource(data=dict(
    act=['0', '1'],
    hit=[target.group.value_counts()['TN'],
         target.group.value_counts()['TP']],
    miss=[target.group.value_counts()['FP'],
          target.group.value_counts()['FN']]))
p_bar = figure(x_range=['0', '1'], plot_width=380, plot_height=100,
               title='')
p_bar.vbar_stack(['hit', 'miss'], x='act', width=0.3,
                 color=['steelblue', 'red'], source=bar_data, alpha=0.7,
                 legend=['Hit', 'Miss'])
p_bar.y_range.start = 0
p_bar.x_range.range_padding = 0.1
p_bar.xgrid.grid_line_color = None
p_bar.axis.minor_tick_line_color = None
p_bar.outline_line_color = None
p_bar.legend.location = "center"
p_bar.xaxis.axis_label = 'Actual Label'
p_bar.yaxis.axis_label = 'Frequency'


# Histogram
hist, edges = np.histogram(target.scores, bins=30)
hist_data = ColumnDataSource(data=dict(scores=hist, left=edges[:-1],
                                       right=edges[1:]))
p_hist = figure(plot_width=380, plot_height=160)
p_hist.quad(bottom=0, top='scores', left='left', right='right',
            source=hist_data, fill_color='steelblue', line_color='white')
p_hist.title.text_font_size = "20px"
p_hist.xaxis.ticker = FixedTicker(ticks=np.arange(0, 1.1, 0.1))
p_hist.xaxis.axis_label = 'Scores'
p_hist.yaxis.axis_label = 'Frequency'


# Data input
text_input = TextInput(value='', title='Data path: ')
slider = Slider(title='Threshold', start=0, end=1, value=0.7, step=0.01)
error_msg = PreText(text='')
roc_msg = PreText(text='Optimal Threshold by ROC: ', width=100, height=20)
pr_msg = PreText(text='Optimal Threshold by PR: ', width=100, height=20)
download = ColumnDataSource(data=dict(
    scores=target.scores,
    group=target.group))


def update_data(attrname, old, new):
    try:
        df = pd.read_csv(text_input.value)
        if len(df) > 5000:
            df = df.sample(n=5000, random_state=0)
        scores = np.array(df.scores)
        labels = np.array(df.labels)
        val = slider.value
        hline.data = dict(x=[-0.3, 1.3], y=[val, val])
        target = classify(scores, labels, val)

        # Scatterplot
        scatter = target.sample(n=500, random_state=0)
        TP.data = dict(x=scatter['position'][scatter.group == 'TP'],
                       y=scatter['scores'][scatter.group == 'TP'])
        TN.data = dict(x=scatter['position'][scatter.group == 'TN'],
                       y=scatter['scores'][scatter.group == 'TN'])
        FP.data = dict(x=scatter['position'][scatter.group == 'FP'],
                       y=scatter['scores'][scatter.group == 'FP'])
        FN.data = dict(x=scatter['position'][scatter.group == 'FN'],
                       y=scatter['scores'][scatter.group == 'FN'])

        # ROC
        fpr, tpr, thresholds = roc_curve(target.actual_label, target.scores)
        tnr = 1 - fpr
        tf = tpr - tnr
        optimal_threshold = thresholds[abs(tf).argsort()[0]]
        roc_msg.text = 'Optimal Threshold by ROC: ' + str(round(
                       optimal_threshold, 2))
        diff = thresholds - val
        x_coord = fpr[abs(diff).argsort()[0]]
        y_coord = tpr[abs(diff).argsort()[0]]
        roc_vline.data = dict(x=[x_coord, x_coord], y=[0, y_coord])
        roc_hline.data = dict(x=[x_coord, 1], y=[y_coord, y_coord])
        roc.data = dict(x=fpr, y=tpr)
        label_tpr.text = 'TPR: ' + str(round(y_coord, 3))
        label_tnr.text = 'TNR: ' + str(round(1 - x_coord, 3))

        # PR
        precision, recall, thresholds_pr = precision_recall_curve(labels,
                                                                  scores)
        thresholds_pr = np.append(thresholds_pr, 1)
        pr.data = dict(x=recall, y=precision)
        diff_pr = thresholds_pr - val
        tf_pr = precision - recall
        optimal_threshold_pr = thresholds_pr[abs(tf_pr).argsort()[0]]
        pr_msg.text = 'Optimal Threshold by PR: ' + str(round(
                      optimal_threshold_pr, 2))
        x_coord_pr = recall[abs(diff_pr).argsort()[0]]
        y_coord_pr = precision[abs(diff_pr).argsort()[0]]
        pr_vline.data = dict(x=[x_coord_pr, x_coord_pr], y=[0, y_coord_pr])
        pr_hline.data = dict(x=[0, x_coord_pr], y=[y_coord_pr, y_coord_pr])
        label_pre.text = 'Precision: ' + str(round(y_coord_pr, 3))
        label_rec.text = 'Recall: ' + str(round(x_coord_pr, 3))

        # Bar plot
        bar_data.data = dict(act=['0', '1'],
                             hit=[target.group.value_counts()['TN'],
                                  target.group.value_counts()['TP']],
                             miss=[target.group.value_counts()['FP'],
                                   target.group.value_counts()['FN']])

        # Histogram
        hist, edges = np.histogram(target.scores, bins=30)
        hist_data.data = dict(scores=hist, left=edges[:-1], right=edges[1:])

        # Download
        download.data = dict(scores=target.scores, group=target.group)

        # Error message
        error_msg.text = ''

    except FileNotFoundError:
        error_msg.text = 'File does not exist.'


text_input.on_change('value', update_data)
slider.on_change('value', update_data)
button = Button(label="Download", button_type="success")
button.callback = CustomJS(args=dict(source=download), code=open(
                           join(dirname(__file__), "download.js")).read())


grid = gridplot([[text_input, slider],
                 [column([error_msg, roc_msg, pr_msg, button]),
                 p_scatter, p_roc],
                 [None, column(p_bar, p_hist), p_pr]])

curdoc().add_root(grid)
